import warnings


def read(fn, warn=False):
    from preconvert.output import json

    if not isinstance(fn, str):
        return json.load(fn)
    if fn.endswith(".jsonl"):
        if warn:
            warnings.warn("Reading streaming format at once.")
        return list(iread(fn))
    with open(fn) as f:
        return json.load(f)


def append(obj, fn):
    from preconvert.output import json

    if not isinstance(fn, str):
        raise TypeError("Cannot append to compression")
    with open(fn, "a+") as f:
        f.write(json.dumps(obj) + "\n")


def write(obj, fn):
    from preconvert.output import json

    if not isinstance(fn, str):
        fn.write(bytes(json.dumps(obj), encoding="utf8"))
    else:
        with open(fn, "w") as f:
            json.dump(obj, f, indent=4)


def iread(fn):
    from preconvert.output import json

    if not isinstance(fn, str):
        raise TypeError("Cannot iteratively read compressed file now")
    with open(fn) as f:
        for i, line in enumerate(f):
            try:
                yield json.loads(line)
            except Exception as e:
                msg = "JSON-L parsing error in line number {} in the jsonl file".format(i)
                raise Exception(msg, line)


def iwrite(obj, fn):
    from preconvert.output import json

    if not isinstance(fn, str):
        raise TypeError("Cannot iteratively write compressed")
    with open(fn, "w") as f:
        for chunk in obj:
            f.write(json.dumps(chunk) + "\n")



import os
import just.txt as txt
import just.json_ as json
import just.newl as newl
import just.yaml_ as yaml
import just.bytes as bytes_
import just.pickle_ as pickle
from just.path_ import remove
from just import make_path, glob, mkdir
import lxml.html

EXT_TO_MODULE = {
    "html": txt,
    "py": txt,
    "txt": txt,
    "log": txt,
    "sql": txt,
    "md": txt,
    "newl": newl,
    "json": json,
    "jsonl": json,
    "yaml": yaml,
    "yml": yaml,
    "pickle": pickle,
    "pkl": pickle,
    "bytes": bytes_,
    "png": bytes_,
}

EXT_TO_COMPRESSION = {}

try:
    import gzip

    EXT_TO_COMPRESSION.update({"gz": gzip.GzipFile, "gzip": gzip.GzipFile})
except ImportError:
    pass

try:
    import bz2

    EXT_TO_COMPRESSION.update({"bz2": bz2.BZ2File, "bzip2": bz2.BZ2File, "bzip": bz2.BZ2File})
except ImportError:
    pass

try:
    import lzma

    EXT_TO_COMPRESSION.update({"xz": lzma.open})
except ImportError:
    pass

try:
    import just.zstd_ as zstd

    EXT_TO_COMPRESSION.update({"zstd": zstd.ZstdFile, "zst": zstd.ZstdFile, "zstandard": zstd.ZstdFile})
except ImportError:
    pass

ZSTD_AVAILABLE = "zstd" in EXT_TO_COMPRESSION
ROLL_EXTENSION = "zstd" if ZSTD_AVAILABLE else "gz"


def reader(fname, no_exist, read_func_name, unknown_type, ignore_exceptions):
    fname = make_path(fname)
    if not os.path.isfile(fname) and no_exist is not None:
        return no_exist
    compression = []
    stripped_fname = fname
    for k, v in EXT_TO_COMPRESSION.items():
        if fname.endswith(k):
            compression.append(v)
            stripped_fname = stripped_fname[: -(len(k) + 1)]
    ext = stripped_fname.split(".")[-1] if "." in stripped_fname[-6:] else None
    if ext not in EXT_TO_MODULE and unknown_type == "RAISE":
        raise TypeError("just does not yet cover '{}'".format(ext))
    reader_module = EXT_TO_MODULE.get(ext, None) or EXT_TO_MODULE[unknown_type]
    read_fn = getattr(reader_module, read_func_name)
    if ignore_exceptions is not None:
        try:
            if compression:
                compression = compression[0]
                # actually returns a file handler >.<
                with compression(fname, "rb") as f:
                    return read_fn(f)
            else:
                return read_fn(fname)
        except ignore_exceptions:
            return None
    else:
        if compression:
            compression = compression[0]
            # actually returns a file handler >.<
            with compression(fname, "rb") as f:
                return read_fn(f)
        else:
            return read_fn(fname)


def read(fname, no_exist=None, unknown_type="RAISE", ignore_exceptions=None):
    return reader(fname, no_exist, "read", unknown_type, ignore_exceptions)


def multi_read(star_path, no_exist=None, unknown_type="RAISE", ignore_exceptions=None, sort_reverse=False):
    for x in glob(star_path, sort_reverse=sort_reverse):
        yield x, read(x, no_exist, unknown_type, ignore_exceptions)


def writer(obj, fname, mkdir_no_exist, skip_if_exist, write_func_name, unknown_type):
    fname = make_path(fname)
    if skip_if_exist and os.path.isfile(fname):  # pragma: no cover
        return False
    if mkdir_no_exist:
        dname = os.path.dirname(fname)
        if dname not in set([".", "..", ""]):
            mkdir(dname)
    compression = []
    stripped_fname = fname
    for k, v in EXT_TO_COMPRESSION.items():
        if fname.endswith(k):
            compression.append(v)
            stripped_fname = stripped_fname[: -(len(k) + 1)]

    ext = stripped_fname.split(".")[-1] if "." in stripped_fname[-6:] else None
    if ext not in EXT_TO_MODULE and unknown_type == "RAISE":
        raise TypeError("just does not yet cover '{}'".format(ext))
    writer_module = EXT_TO_MODULE.get(ext, None) or EXT_TO_MODULE[unknown_type]
    write_fn = getattr(writer_module, write_func_name)
    if compression:
        # actually returns a file handler >.<
        compression = compression[0]
        with compression(fname, "wb") as f:
            return write_fn(obj, f)
    else:
        return write_fn(obj, fname)


def write(obj, fname, mkdir_no_exist=True, skip_if_exist=False, unknown_type="RAISE"):
    return writer(obj, fname, mkdir_no_exist, skip_if_exist, "write", unknown_type)


# only supported for JSON Lines so far.
roll_counter = {}


def get_rolled_path(fname, roll_extension):
    it = 0
    dir_, base = os.path.split(fname)
    base, *rest = base.split(".")
    while True:
        path = f"{dir_}/{base}_{it}.{'.'.join(rest)}.{roll_extension}"
        if not os.path.exists(path):
            return path
        it += 1


def append(
    obj,
    fname,
    mkdir_no_exist=True,
    skip_if_exist=False,
    unknown_type="RAISE",
    roll=0,
    roll_extension=ROLL_EXTENSION,
):
    writer(obj, fname, mkdir_no_exist, skip_if_exist, "append", unknown_type)
    if roll:
        fname = make_path(fname)
        if fname not in roll_counter:
            roll_counter[fname] = 0
        roll_counter[fname] += 1
        if roll == roll_counter[fname]:
            data = read(fname)
            rolled_path = get_rolled_path(fname, roll_extension)
            write(data, rolled_path, unknown_type=unknown_type)
            remove(fname)
            roll_counter[fname] = 0


def multi_write(obj, fname, mkdir_no_exist=True, skip_if_exist=False):
    if not isinstance(fname, list) or not isinstance(obj, list):  # pragma: no cover
        raise NotImplementedError("Only list of fnames + list of objects supported.")
    return [write(o, fn, mkdir_no_exist, skip_if_exist) for o, fn in zip(obj, fname)]


def iread(fname, no_exist=None, unknown_type="RAISE", ignore_exceptions=None):
    return reader(fname, no_exist, "iread", unknown_type, ignore_exceptions)


def iwrite(obj, fname, mkdir_no_exist=True, skip_if_exist=False, unknown_type="RAISE"):
    return writer(obj, fname, mkdir_no_exist, skip_if_exist, "iwrite", unknown_type)


def read_tree(fname):
    return lxml.html.fromstring(read(fname, unknown_type="html"))



import os
import errno


def mkdir(path, mode=0o777):
    path = os.path.expanduser(path)
    try:
        os.makedirs(path, mode)
    # Python >2.5
    except OSError as exc:  # pragma: no cover
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise
    return path



"""
This module implements a set of requests TransportAdapter, PoolManager,
ConnectionPool and HTTPSConnection with one goal only:
* to use a specific IP address when connecting via SSL to a web service without
running into SNI trouble.
The usual technique to force an IP address on an HTTP connection with Requests
is (assuming I want http://example.com/some/path on IP 1.2.3.4):
requests.get("http://1.2.3.4/some/path", headers={'Host': 'example.com'})
this is useful if I want to specifically test how 1.2.3.4 is responding; for
instance, if example.com is DNS round-robined to several IP
addresses and I want to hit one of them specifically.
This also works for https requests if using Python <2.7.9 because older
versions don't do SNI and thus don't pass the requested hostname as part of the
SSL handshake.
However, Python >=2.7.9 and >=3.4.x conveniently added SNI support, breaking
this way of connecting to the IP, because the IP address embedded in the URL
*is* passed as part of the SSL handshake, causing errors (mainly, the server
returns a 400 Bad Request because the SNI host 1.2.3.4 doesn't match the one in
the HTTP headers example.com).
The "easiest" way to achieve this is to force the IP address at the lowest
possible level, namely when we do socket.create_connection. The rest of the
"stack" is given the actual hostname. So the sequence is:
1- Open a socket to 1.2.3.4
2- SSL wrap this socket using the hostname.
3- Do the rest of the HTTPS traffic, headers and all over this socket.
Unfortunately Requests hides the socket.create_connection call in the deep
recesses of urllib3, so the specified chain of classes is needed to propagate
the given dest_ip value all the way down the stack.
Because this applies to a very limited set of circumstances, the overridden
code is very simplistic and eschews many of the nice checks Requests does for
you.
Specifically:
- It ONLY handles HTTPS.
- It does NO certificate verification (which would be pointless)
- Only tested with Requests 2.2.1 and 2.9.1.
- Does NOT work with the ancient urllib3 (1.7.1) shipped with Ubuntu 14.04.
  Should not be an issue because Ubunt 14.04 has older Python which doesn't do
  SNI.
How to use it
=============
It's like any other transport adapter. Just pass the IP address that
connections to the given URL prefix should use.
session = requests.Session()
session.mount("https://example.com", ForcedIPHTTPSAdapter(dest_ip='1.2.3.4'))
response = session.get(
    '/some/path', headers={'Host': 'example.com'}, verify=False)
"""
# Note this module will ImportError if there's no sane requests/urllib
# combination available so the adapter won't work, and it's up to the caller to
# decide what to do. The caller can, for instance, check the Python version and
# if it's <2.7.9 decide to use the old "http://$IP/ technique. If Python is
# >2.7.9 and the adapter doesn't work, unfortunately, there's nothing that can
# be done :(
from distutils.version import StrictVersion
from socket import error as SocketError, timeout as SocketTimeout

import requests
from requests.adapters import HTTPAdapter

from requests.packages.urllib3.poolmanager import (
    PoolManager,
    HTTPSConnectionPool,
)
from requests.packages.urllib3.exceptions import ConnectTimeoutError

try:
    # For requests 2.9.x
    from requests.packages.urllib3.util import connection
    from requests.packages.urllib3.exceptions import NewConnectionError
except ImportError:
    # For requests  <= 2.2.x
    import socket as connection
    from socket import error as NewConnectionError
# Requests older than 2.4.0's VerifiedHHTPSConnection is broken and doesn't
# properly use _new_conn. On these versions, use UnverifiedHTTPSConnection
# instead.
if StrictVersion(requests.__version__) < StrictVersion("2.4.0"):
    from requests.packages.urllib3.connection import UnverifiedHTTPSConnection as HTTPSConnection
else:
    from requests.packages.urllib3.connection import HTTPSConnection


class ForcedIPHTTPSAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.dest_ip = kwargs.pop("dest_ip", None)
        super(ForcedIPHTTPSAdapter, self).__init__(*args, **kwargs)

    def init_poolmanager(self, *args, **pool_kwargs):
        pool_kwargs["dest_ip"] = self.dest_ip
        self.poolmanager = ForcedIPHTTPSPoolManager(*args, **pool_kwargs)


class ForcedIPHTTPSPoolManager(PoolManager):
    def __init__(self, *args, **kwargs):
        self.dest_ip = kwargs.pop("dest_ip", None)
        super(ForcedIPHTTPSPoolManager, self).__init__(*args, **kwargs)

    def _new_pool(self, scheme, host, port, request_context=None):
        assert scheme == "https"
        kwargs = self.connection_pool_kw.copy()
        kwargs["dest_ip"] = self.dest_ip
        return ForcedIPHTTPSConnectionPool(host, port, **kwargs)


class ForcedIPHTTPSConnectionPool(HTTPSConnectionPool):
    def __init__(self, *args, **kwargs):
        self.dest_ip = kwargs.pop("dest_ip", None)
        super(ForcedIPHTTPSConnectionPool, self).__init__(*args, **kwargs)

    def _new_conn(self):
        self.num_connections += 1

        actual_host = self.host
        actual_port = self.port
        if self.proxy is not None:
            actual_host = self.proxy.host
            actual_port = self.proxy.port

        conn_kw = getattr(self, "conn_kw", {}).copy()
        conn_kw["dest_ip"] = self.dest_ip
        conn = ForcedIPHTTPSConnection(
            host=actual_host, port=actual_port, timeout=self.timeout.connect_timeout, strict=self.strict, **conn_kw
        )
        pc = self._prepare_conn(conn)
        return pc

    def __str__(self):
        return "%s(host=%r, port=%r, dest_ip=%s)" % (type(self).__name__, self.host, self.port, self.dest_ip)


class ForcedIPHTTPSConnection(HTTPSConnection, object):
    def __init__(self, *args, **kwargs):
        self.dest_ip = kwargs.pop("dest_ip", None)
        super(ForcedIPHTTPSConnection, self).__init__(*args, **kwargs)

    def _new_conn(self):
        extra_kw = {}
        if self.source_address:
            extra_kw["source_address"] = self.source_address

        if getattr(self, "socket_options", None):
            extra_kw["socket_options"] = self.socket_options

        dest_host = self.dest_ip if self.dest_ip else self.host

        try:
            conn = connection.create_connection((dest_host, self.port), self.timeout, **extra_kw)

        except SocketTimeout as e:
            raise ConnectTimeoutError(
                self, "Connection to %s timed out. (connect timeout=%s)" % (self.host, self.timeout)
            )

        except SocketError as e:
            raise NewConnectionError(self, "Failed to establish a new connection: %s" % e)

        return conn



import os
import time
import warnings
import hashlib
from collections import defaultdict

from just.dir import mkdir
from just.path_ import exists, glob, make_path, remove, rename
from just.read_write import write, read
from just.forcedip import ForcedIPHTTPSAdapter
from just.source_ip import SourceAddressAdapter


session = None
PER_FOLDER = 3000

caches = {}
timers = {}
sessions = {}
last_cache_fname = {}

obj_counts = defaultdict(int)


def get_domain(url):
    return url.split("/")[2].split("?")[0].replace("www.", "")


def delete_from_cache(url, cache_key=None, compression=".gz"):
    domain_name = get_domain(url)
    if cache_key is not None:
        fname = get_cache_file_name_str(domain_name, cache_key, compression)
        parts = fname.split()
        parts[-2] = "*"
        fname = glob("/".join(parts))[0]
    else:
        fname = last_cache_fname[domain_name]
    remove(fname)
    return fname


def update_obj_count(domain, dir_name, obj_type):
    if (domain, obj_type) not in obj_counts:
        dir_name = make_path(dir_name)
        try:
            existing_partitions = glob(make_path(dir_name) + "/*")
            if existing_partitions:
                latest_folder = max(existing_partitions)
                base = int(latest_folder.split("/")[-1]) * PER_FOLDER
                count = base + len(os.listdir(latest_folder))
            else:
                count = 0
        except FileNotFoundError:
            count = 0
        obj_counts[(domain, obj_type)] = count
    return obj_counts[(domain, obj_type)]


def get_cache_file_name(domain, request_info, compression=".gz"):
    from preconvert.output import json

    key = json.dumps(request_info)
    m = hashlib.md5()
    m.update(key.encode("utf8"))
    md5 = m.hexdigest()
    dir_name, fname = md5[:3], md5[3:]
    if not compression:
        compression = ""
    return f"~/.just_requests/{domain}/{dir_name}/{fname}.json{compression}"


def get_obj_type(cache_key):
    if isinstance(cache_key, bool):
        obj_type, fname = "", ""
    else:
        if "/" in cache_key:
            obj_type, fname = cache_key.split("/")
        else:
            obj_type, fname = "", cache_key
        if obj_type:
            obj_type = "/" + obj_type
    return obj_type, fname


def get_cache_file_name_str(domain, cache_key, compression=".gz"):
    obj_type, fname = get_obj_type(cache_key)
    initial_part = f"~/.just_requests/{domain}{obj_type}"
    partition = update_obj_count(domain, initial_part, obj_type) // PER_FOLDER
    if not compression:
        compression = ""
    return f"{initial_part}/{partition}/{fname}.json{compression}"


def from_cache(url, cache_key, paths_only=False, compression=".gz"):
    domain = get_domain(url)
    obj_type, fname = get_obj_type(cache_key)
    initial_part = f"~/.just_requests/{domain}{obj_type}"
    partition = "*"
    paths = glob(f"{initial_part}/{partition}/{fname}.json{compression}")
    if paths_only:
        return paths
    return [read(x)["resp"] for x in paths]


def _retry(
    method,
    max_retries,
    delay_base,
    raw,
    caching,
    cache_compression,
    sleep_time,
    reuse_session,
    local_address,
    remote_ip,
    kwargs,
):
    import requests
    from requests import RequestException, Session
    from requests.utils import cookiejar_from_dict

    tries = 0
    url = kwargs["url"]
    domain_name = get_domain(url)

    use_cache, *request_info = caching

    cache_file_name = get_cache_file_name(domain_name, request_info, cache_compression)

    if exists(cache_file_name):
        if use_cache:
            last_cache_fname[domain_name] = cache_file_name
            return read(cache_file_name)["resp"]
        if use_cache is None:
            use_cache = True
        remove(cache_file_name)

    if "timeout" not in kwargs:
        kwargs["timeout"] = delay_base

    cookies = kwargs.get("cookies")
    if isinstance(cookies, dict):
        kwargs["cookies"] = cookiejar_from_dict(cookies)

    session_key = (domain_name, local_address, remote_ip)
    if reuse_session:
        t1 = time.time()
        expired = False
        if session_key not in sessions or (expired := sessions[session_key][1] + 300 < t1):
            session = Session()
            if local_address is not None:
                session.mount("http://", SourceAddressAdapter(local_address))
                session.mount("https://", SourceAddressAdapter(local_address))
            if remote_ip is not None:
                session.mount(f"https://{domain_name}", ForcedIPHTTPSAdapter(dest_ip=remote_ip))
            sessions[session_key] = [session, t1]

        if expired:
            print("just.requests_", kwargs["url"], "old session")

        # e.g. GET or POST
        request_fn = getattr(sessions[session_key][0], method)
    else:
        request_fn = getattr(requests, method)

    if sleep_time and session_key in timers:
        # 1200 - 1201 + 3
        diff = timers[session_key] - time.time() + sleep_time

        if diff > 0:
            time.sleep(diff)

    # retrying
    err = False
    r = None
    while tries < max_retries:
        try:
            r = request_fn(**kwargs)
            # update session age
            if reuse_session:
                sessions[session_key][1] = time.time()
            if r.status_code > 399:
                err = None
            break
        except RequestException as e:
            print("just.requests_", kwargs["url"], "attempt", tries, str(e))
            if tries == max_retries:
                err = ""
                r = None
                break
            tries += 1
            time.sleep(delay_base**tries)

    timers[session_key] = time.time()

    # result handling
    if err is None or err == "":
        text = r.text[:500] if r is not None else ""
        if len(text) == 500:
            text += "..."
        code = r.status_code if r is not None else None
        print("ERR", code, url, text)
        tmp = err
        try:
            err = r.json()
        except:
            try:
                err = r.text
            except:
                err = ""
        r = tmp
    elif raw:
        r = r.content
    elif r is None:
        pass
    elif r is not None and "application/json" in r.headers["Content-Type"]:
        r = r.json()
    else:
        r = r.text

    if use_cache and r is not None:
        result = {"resp": r, "request_info": request_info, "response_ts": time.time()}
        if err:
            result["error"] = err
        write(result, cache_file_name)
        obj_type, _ = get_obj_type(use_cache)
        obj_counts[(domain_name, obj_type)] += 1

    return r


def get(
    url,
    params=None,
    max_retries=2,
    delay_base=3,
    raw=False,
    use_cache=False,
    cache_compression=".gz",
    sleep_time=None,
    fname=None,
    reuse_session=True,
    local_address=None,
    remote_ip=None,
    **kwargs,
):
    caching = (use_cache, url, params)

    kwargs["url"] = url
    if params is not None:
        kwargs["params"] = params

    result = _retry(
        "get",
        max_retries,
        delay_base,
        raw,
        caching,
        cache_compression,
        sleep_time,
        reuse_session,
        local_address,
        remote_ip,
        kwargs,
    )

    if fname is not None:
        write(result, fname)

    return result


get.from_cache = from_cache


def warn_cache():
    warnings.warn(
        "This attribute is deprecated as boolean, it still works but advised to make a str key",
        DeprecationWarning,
        stacklevel=2,
    )


def post(
    url,
    params=None,
    data=None,
    max_retries=5,
    raw=False,
    json=None,
    delay_base=3,
    use_cache=False,
    cache_compression=".gz",
    sleep_time=None,
    fname=None,
    reuse_session=True,
    local_address=None,
    remote_ip=None,
    **kwargs,
):
    if isinstance(use_cache, bool):
        warn_cache()

    caching = (use_cache, url, params, data, json)

    kwargs["url"] = url
    if params is not None:
        kwargs["params"] = params
    if data is not None:
        kwargs["data"] = data
    if json is not None:
        kwargs["json"] = json

    result = _retry(
        "post",
        max_retries,
        delay_base,
        raw,
        caching,
        cache_compression,
        sleep_time,
        reuse_session,
        local_address,
        remote_ip,
        kwargs,
    )

    if fname is not None:
        write(result, fname)

    return result


post.from_cache = from_cache


def get_tree(*args, **kwargs):
    import lxml.html

    return lxml.html.fromstring(get(*args, **kwargs))


def post_tree(*args, **kwargs):
    import lxml.html

    return lxml.html.fromstring(post(*args, **kwargs))


def save_session(name, session):
    from requests.utils import dict_from_cookiejar

    if any(["Session" in x.__name__ for x in session.__class__.__mro__]):
        try:
            print("trf")
            session.transfer_driver_cookies_to_session()
        except Exception as e:
            print("ERR", e)
        session = {"headers": session.headers, "cookies": dict_from_cookiejar(session.cookies)}
    write(session, f"~/.just_sessions/" + name + ".json")
